#!/usr/bin/env python3
"""
A very lightweight console menu.
"""

__author__ = "Andreas Ehrlund"
__version__ = "1.3.0"
__license__ = "MIT"

import sys

PADDING = 2
MIN_WIDTH = 32
DELIMITER = ":"
TITLE_INDENT = 3


class Menu:
    def __init__(
        self,
        menu_title,
        message_top="",
        message_bottom="",
        min_width=MIN_WIDTH,
        cycle=True,
    ):
        if not menu_title:
            raise ValueError("Menu title required")

        self.cycle = cycle
        self.min_width = min_width
        self.menu_title = menu_title
        self.message_top = message_top
        self.menu_items = []
        self.message_bottom = [message_bottom] if message_bottom else []

    def add_function_item(self, item_title, func_ref, kwargs):
        self.menu_items.append((item_title, func_ref, kwargs))

    def add_divider(self):
        self.menu_items.append(("-", None))

    def add_message_bottom_row(self, message):
        self.message_bottom.append(message)

    def show(self):
        self.print_menu()

        while True:
            try:
                choice = int(input("Action number: "))
            except ValueError:
                print("Incorrect input, try again.")
                continue

            if 0 == choice:
                return True
                # yield
            elif choice in range(len(self.menu_items) + 1):
                func = self.menu_items[choice - 1][1]
                kwargs = self.menu_items[choice - 1][2]
                func(**kwargs)
                if self.cycle:
                    self.show()
                return False

            else:
                print("Choose a valid item.")

    def get_total_menu_width(self):
        lengths = [
            len((f"{str(ind+1)}{DELIMITER} {item[0]}"))
            for ind, item in enumerate(self.menu_items)
        ]
        lengths.append(TITLE_INDENT + len(self.menu_title))

        if self.message_top:
            lengths.append(len(self.message_top))
        if self.message_bottom:
            lengths.append(len(max(self.message_bottom, key=len)))

        # +4 for border and padding
        return max(max(lengths) + 4, self.min_width)

    def print_menu(self):

        total_menu_width = self.get_total_menu_width()

        menu_top_left = f"╭{TITLE_INDENT*'─'} {self.menu_title} "
        menu_top_right = "╮"
        menu_top_padding = (
            total_menu_width - len(menu_top_left) - len(menu_top_right)
        ) * "─"
        menu_top = f"{menu_top_left}{menu_top_padding}{menu_top_right}"
        print(menu_top)

        # Top message

        if self.message_top:
            print(self.menu_item_string(self.message_top, total_menu_width))
            print("╞" + (total_menu_width - 2) * "═" + "╡")

        # Menu items

        index = 1
        for item in self.menu_items:
            if item[0] == "-" and item[1] == None:
                print("├" + (total_menu_width - 2) * "─" + "┤")
            else:
                print(self.menu_item_string(item[0], total_menu_width, index))
            index += 1

        print(self.menu_item_string("Exit", total_menu_width, 0))

        # Bottom message

        if self.message_bottom:
            print("├" + (total_menu_width - 2) * "─" + "┤")
            for message in self.message_bottom:
                print(self.menu_item_string(message, total_menu_width))

        print("╰" + (total_menu_width - 2) * "─" + "╯")

    def menu_item_string(self, title, total_menu_width, item_no=None):
        index = ""
        if isinstance(item_no, int):
            index = f"{str(item_no)}{DELIMITER} "
        string_left = f"│ {index}{title} "
        string_right = f"│"
        padding_length = total_menu_width - len(string_left) - len(string_right)
        return f"{string_left}{padding_length * ' '}{string_right}"



import pytest
import mock
import builtins
from unittest.mock import patch
from unittest.mock import MagicMock
from io import StringIO

from micromenu import micromenu

# pytest --cov=micromenu --cov-report xml:coverage.xml tests/
# pytest --cov=micromenu tests/


class TestMenu:
    def test_add_function_item(self):
        menu = micromenu.Menu(
            "test", "test messsage top", "test message bottom", min_width=20
        )
        assert len(menu.menu_items) == 0
        menu.add_function_item("title1", lambda x: len(x), {"x": "testparam"})
        assert len(menu.menu_items) == 1

    def test_add_divider(self):
        menu = micromenu.Menu(
            "test", "test messsage top", "test message bottom", min_width=20
        )
        assert len(menu.menu_items) == 0
        menu.add_function_item("title1", lambda x: len(x), {"x": "testparam"})
        menu.add_divider()
        menu.add_function_item("title2", lambda x: len(x), {"x": "testparam"})
        assert len(menu.menu_items) == 3

    def test_add_message_bottom_item_no_init(self):
        menu = micromenu.Menu("test", "test messsage top", min_width=20)
        assert len(menu.menu_items) == 0
        menu.add_message_bottom_row("test message bottom")
        assert len(menu.message_bottom) == 1

    def test_add_message_bottom_item_with_init(self):
        menu = micromenu.Menu(
            "test", "test messsage top", "test message bottom", min_width=20
        )
        assert len(menu.menu_items) == 0
        menu.add_message_bottom_row("test message bottom 2")
        menu.add_message_bottom_row("test message bottom 3")
        assert len(menu.message_bottom) == 3

    def test_print_menu(self, capsys):
        menu = micromenu.Menu("test", "test messsage top", "test message bottom")
        menu.add_function_item("title1", lambda x: len(x), {"x": "testparam"})
        menu.add_divider()
        menu.add_function_item("title2", lambda x: len(x), {"x": "testparam"})
        menu.print_menu()
        captured = capsys.readouterr()
        assert captured.out.startswith("╭─── test ")

    def test_inputs(self):
        with pytest.raises(ValueError) as excinfo:
            menu = micromenu.Menu("")

        menu = micromenu.Menu("test_title", "")
        assert isinstance(menu, micromenu.Menu)

    def test_dynamic_width(self, capsys):
        overflow = 4

        # test menu title longest
        menu = micromenu.Menu("X" * (micromenu.MIN_WIDTH + overflow), "x")
        menu.add_function_item("x", lambda x: len(x), {"x": "testparam"})
        menu.print_menu()
        captured = capsys.readouterr()
        lines = captured.out.split("\n")
        del lines[-1]
        assert all(len(x) == len(lines[0]) for x in lines)

        # test top menu message longest
        menu = micromenu.Menu("x", "X" * (micromenu.MIN_WIDTH + overflow))
        menu.add_function_item("x", lambda x: len(x), {"x": "testparam"})
        menu.print_menu()
        captured = capsys.readouterr()
        lines = captured.out.split("\n")
        del lines[-1]
        assert all(len(x) == len(lines[0]) for x in lines)

        # test bottom menu message longest
        menu = micromenu.Menu("x", "", "X" * (micromenu.MIN_WIDTH + overflow))
        menu.add_function_item("x", lambda x: len(x), {"x": "testparam"})
        menu.add_message_bottom_row("notlongest")
        menu.print_menu()
        captured = capsys.readouterr()
        lines = captured.out.split("\n")
        del lines[-1]
        assert all(len(x) == len(lines[0]) for x in lines)

        # test menu item longest
        def test_item_indexes(num_menu_items):
            menu = micromenu.Menu("x", "x")
            for i in range(num_menu_items):
                menu.add_function_item(
                    "X" * (micromenu.MIN_WIDTH + overflow),
                    lambda x: len(x),
                    {"x": "testparam"},
                )
            menu.print_menu()
            captured = capsys.readouterr()
            lines = captured.out.split("\n")
            del lines[-1]
            assert all(len(x) == len(lines[0]) for x in lines)

        test_item_indexes(9)
        test_item_indexes(10)
        test_item_indexes(11)

    def test_invalid_input(self, capsys):
        menu = micromenu.Menu("test", "test messsage top", "test message bottom")
        menu.add_function_item("title1", lambda x: len(x), {"x": "testparam"})

        with patch("sys.stdin", StringIO("a\n0")):
            menu.show()
        captured = capsys.readouterr()
        assert "Incorrect input, try again." in captured.out

    def test_menu_item_out_of_range(self, capsys):
        menu = micromenu.Menu("test", "test messsage top", "test message bottom")
        menu.add_function_item("title1", lambda x: len(x), {"x": "testparam"})

        with patch("sys.stdin", StringIO("9\n0")):
            menu.show()
        captured = capsys.readouterr()
        assert "Choose a valid item." in captured.out

    def test_menu_item_function_called(self, capsys):
        dummy = MagicMock()

        menu = micromenu.Menu("test", "test messsage top", "test message bottom")
        menu.add_function_item("title1", dummy, {"x": "testparam"})

        with patch("sys.stdin", StringIO("1\n0\n0")):
            menu.show()
            assert dummy.called

    def test_menu_terminates(self, capsys):
        menu = micromenu.Menu("test", cycle=False)
        menu.add_function_item("title1", lambda x: print(x), {"x": "done"})

        with patch("sys.stdin", StringIO("0\n0")):
            assert menu.show() == True
